export interface UdfArgument {
    type: string;
    multiple?: boolean;
    keywords?: string[];
    optional?: boolean;
}
export interface UdfDetails {
    returnTypes: string[];
    name: string;
    arguments: UdfArgument[][];
    altArguments?: UdfArgument[][];
    signature: string;
    draggable: string;
    description?: string;
    described?: boolean;
}
export interface UdfCategoryFunctions {
    [attr: string]: UdfDetails;
}
export interface UdfCategory {
    name: string;
    functions: UdfCategoryFunctions;
    isAnalytic?: boolean;
    isAggregate?: boolean;
}
export interface SetOptions {
    [attr: string]: SetDetails;
}
export interface SetDetails {
    default: string;
    type: string;
    description: string;
}
export interface TypeConversion {
    [attr: string]: {
        [attr: string]: boolean;
    };
}
export interface SqlReferenceProvider {
    getReservedKeywords(dialect: string): Promise<Set<string>>;
    getSetOptions(dialect: string): Promise<SetOptions>;
    getUdfCategories(dialect: string): Promise<UdfCategory[]>;
    hasUdfCategories(dialect: string): boolean;
}
/**
 * Matches types based on implicit conversion i.e. if you expect a BIGINT then INT is ok but not BOOLEAN etc.
 */
export declare function matchesType(_dialect: string, expectedTypes: string[], actualRawTypes: string[]): boolean;
